#include <block_cache.h>
#include <exception.h>
#include <hdd.h>
#include <kmalloc.h>
#include <utils.h>

/**
 * Resetea un vector de cache a sus valores defaults
 */
static void block_cache_reset_vector(block_cache_entry* vec, uint64_t count) {
    // El valor default del vector de la cache es 0 para todas las entradas
    memset(vec, 0, sizeof(block_cache_entry) * count);
}

/**
 * Flushea un bloque a disco
 * @param entry_number Posicion en el vector del bloque a flushear
 */
static void block_cache_flush_entry(block_cache* cache, uint64_t entry_number) {
    block_cache_entry* vec = cache->vec;
    uint64_t block_number = vec[entry_number].number;
    hdd_write(block_number * cache->size_sectors, cache->size_sectors,
              vec[entry_number].data);
    kfree(vec[entry_number].data);
    vec[entry_number].data = NULL;
    vec[entry_number].flags &= ~BLOCK_CACHE_DIRTY;
}

/**
 * Duplica el tamaño de la cache, util si nos quedamos sin espacio
 * @returns La primera entrada nueva
 */
static block_cache_entry* block_cache_expand(block_cache* cache) {
    uint64_t current_size = cache->count * sizeof(block_cache_entry);
    // Guardamos el puntero al vector anterior
    block_cache_entry* old_vec = cache->vec;
    // Hacemos espacio para el nuevo
    cache->vec = kmalloc(current_size * 2);
    // Copiamos el vector viejo
    memcpy(cache->vec, old_vec, current_size);
    // La posicion de la primera nueva entrada
    block_cache_entry* new_entry = cache->vec + cache->count;
    // Reseteamos el resto
    block_cache_reset_vector(new_entry, cache->count);
    // Actualizamos el count
    cache->count *= 2;
    // Liberamos el vector viejo
    kfree(old_vec);
    // Retornamos la nueva entrada
    return new_entry;
}

/**
 * Crea una cache de bloques de tamaño definido con una cantidad inicial
 */
block_cache* block_cache_create(uint64_t block_size, uint64_t count) {
    block_cache* cache = kmalloc(sizeof(block_cache));
    cache->vec = kmalloc(sizeof(block_cache_entry) * count);
    // Division entera con techo
    uint64_t ssize =
        block_size / ATA_SECTOR_SIZE + (block_size % ATA_SECTOR_SIZE != 0);
    cache->size_sectors = ssize;
    cache->size_bytes = ssize * block_size;
    cache->count = count;
    block_cache_reset_vector(cache->vec, cache->count);
    return cache;
}

/**
 * Carga un bloque en la cache, despues hay que liberarlo con unload_block
 * Si no hay espacio la expande
 * WARNING: No es thread safe
 * @param cache La cache a donde cargar el bloque
 * @param block_number El numero de bloque absoluto a cargar en cache
 * @returns Puntero directo al bloque cacheado
 */
void* block_cache_load(block_cache* cache, uint64_t block_number) {
    // Buscamos la entrada y mientras nos fijamos si hay alguno libre
    block_cache_entry* last = cache->vec + cache->count;
    block_cache_entry* free_entry = NULL;
    for (block_cache_entry* entry = cache->vec; entry != last; entry++) {
        if (entry->number == block_number && entry->data) {
            entry->ref_count++;
            return entry->data;
        }
        // Priorizamos los que no tienen alloc
        if ((!free_entry && entry->ref_count == 0) || !entry->data)
            free_entry = entry;
    }
    // Si no hay espacio y no encontramos nada necesitamos expandirla
    if (!free_entry)
        free_entry = block_cache_expand(cache);
    // Si no hay alloc, lo hacemos
    if (!free_entry->data)
        free_entry->data = kmalloc(cache->size_bytes);
    // Actualizamos la cache
    free_entry->number = block_number;
    free_entry->ref_count++;
    // Leemos el bloque
    hdd_read(block_number * cache->size_sectors, cache->size_sectors,
             free_entry->data);
    // Devolvemos el puntero al bloque
    return free_entry->data;
}

/**
 * Sets the DIRTY flag
 * @param cache The cache where it will look for block_ptr
 * @param block_ptr The pointer to the block we want to set the DIRTY flag to
 */
void block_cache_set_dirty(block_cache* cache, void* block_ptr) {
    if (block_ptr == NULL)
        return;
    block_cache_entry* vec = cache->vec;
    // The index
    uint64_t i = 0;
    // We look it up and set the DIRTY flag on
    uint64_t count = cache->count;
    uintptr_t block_int = (uintptr_t)block_ptr;
    while (i < count) {
        // If it is any address inside the block
        if (block_int >= (uintptr_t)vec[i].data &&
            block_int < ((uintptr_t)vec[i].data + cache->size_bytes)) {
            vec[i].flags |= BLOCK_CACHE_DIRTY;
            break;
        }
        i++;
    }

    if (i >= cache->count) {
        KERNEL_PANIC("Direccion %p no es valida en ningun bloque", block_ptr);
    }
}

/**
 * Carga un bloque en la cache y lo setea DIRTY, despues hay que liberarlo
 * con unload_block. Si no hay espacio la expande
 * WARNING: No es thread safe
 * @param cache La cache a donde cargar el bloque
 * @param block_number El numero de bloque absoluto a cargar en cache
 * @returns Puntero directo al bloque cacheado
 */
void* block_cache_load_dirty(block_cache* cache, uint64_t block_number) {
    void* block_ptr = block_cache_load(cache, block_number);
    block_cache_set_dirty(cache, block_ptr);
    return block_ptr;
}

/**
 * Libera el bloque con el puntero que se pasa por parametro
 * @param cache La cache a liberar el bloque
 * @param block_ptr Puntero que apunte al bloque en cualquiera de sus posiciones
 */
void block_cache_unload(block_cache* cache, void* block_ptr) {
    if (block_ptr == NULL)
        return;
    // Tomamos el vector
    block_cache_entry* vec = cache->vec;
    // Nuestro indice
    uint64_t i = 0;
    // Lo buscamos si lo encontramos le sacamos una referencia
    uint64_t count = cache->count;
    uintptr_t block_int = (uintptr_t)block_ptr;
    while (i < count) {
        // Si es cualquier direccion dentro del bloque, por ej un inodo
        if (block_int >= (uintptr_t)vec[i].data &&
            block_int < ((uintptr_t)vec[i].data + cache->size_bytes) &&
            vec[i].ref_count > 0) {
            vec[i].ref_count--;
            break;
        }
        i++;
    }

    if (i < cache->count) {
        // Si la referencia llega a 0 y esta sucio lo flusheamos
        if ((vec[i].ref_count == 0) && (vec[i].flags & BLOCK_CACHE_DIRTY))
            block_cache_flush_entry(cache, i);
    } else {
        KERNEL_PANIC("Direccion %p no es valida en ningun bloque", block_ptr);
    }
}

/**
 * Flushea toda la cache que estaba sucia
 */
void block_cache_flush(block_cache* cache) {
    uint64_t i = 0;
    block_cache_entry* vec = cache->vec;
    while (i < cache->count) {
        if (vec[i].data && (vec[i].flags & BLOCK_CACHE_DIRTY))
            block_cache_flush_entry(cache, i);
        i++;
    }
}

/**
 * Hace free de todos los bloques de la cache
 */
void block_cache_free_blocks(block_cache* cache) {
    // Primero flusheamos todo
    block_cache_flush(cache);
    uint64_t i = 0;
    block_cache_entry* vec = cache->vec;
    while (i < cache->count) {
        if (vec[i].data)
            kfree(vec[i].data);
        i++;
    }
    block_cache_reset_vector(vec, cache->count);
}

/**
 * Hace free de la cache entera incluyendo el vector
 */
void block_cache_free(block_cache* cache) {
    if (cache->vec == NULL)
        return;
    block_cache_free_blocks(cache);
    kfree(cache->vec);
    kfree(cache);
}
