#ifndef __FS_H
#define __FS_H

#include <types.h>

#define MAX_FILENAME_SIZE 256

// Lo define el filesystem
typedef struct fs_file fs_file;
typedef struct fs_dir fs_dir;
typedef struct fs_info fs_info;

typedef struct dir_entry {
    uint8_t  file_type;               // Tipo de archivo.
    char name[MAX_FILENAME_SIZE];     // 255 + 1 por el caracter nulo
} dir_entry;

// TODO: Faltan campos 
typedef struct fs_stat {
    uint64_t size;
    uint16_t mode;
} fs_stat;

#define IS_MODE_TYPE(m,t) (((uint32_t) (m) >> 12) == (t))

#define FILE_ACCESS_STICKY_BIT    0x1
#define FILE_ACCESS_SET_GROUP_ID  0x2
#define FILE_ACCESS_SET_USER_ID   0x4

#define FILE_TYPE_NULL            0x0
#define FILE_TYPE_FIFO            0x1
#define FILE_TYPE_CHAR_DEVICE     0x2
#define FILE_TYPE_DIRECTORY       0x4
#define FILE_TYPE_BLOCK_DEVICE    0x6
#define FILE_TYPE_REGULAR_FILE    0x8
#define FILE_TYPE_SYMBOLIC_LINK   0xA
#define FILE_TYPE_UNIX_SOCKET     0xC

/*
 * Funciones a implementar por un filesystem
 * Deberían estar libres de kmalloc
 */
// Monta el filesystem
int64_t fs_mount(fs_info* info);
// Desmonta el filesystem
void fs_unmount(fs_info* info);
// Devuelve el directorio raiz del filesystem en dir
fs_dir* fs_root(fs_info* info, fs_dir* dir);
// Dado un directorio src abre el archivo apuntado por el en dest
fs_file* fs_open(fs_info* info, fs_dir* src, fs_file* dest);
// Cierra el archivo file
int64_t fs_close(fs_info* info, fs_file* file);
// Lee de un archivo "size" bytes y los pone en "buf"
uint64_t fs_read(fs_info* info, fs_file* file, void* buf, uint64_t size);
// Escribe en un archivo "size" bytes desde "buf"
uint64_t fs_write(fs_info* info, fs_file* file, const void* buf, uint64_t size);
// Avanza offset en el archivo, puede retroceder tambien
int64_t fs_seek(fs_info* info, fs_file* file, int64_t offset); 
// Indica donde esta parado el archivo
int64_t fs_tell(fs_info* info, fs_file* file);
// Rellena stat con informacion del archivo
int64_t fs_fstat(fs_info* info, fs_file* file, fs_stat* stat);
// Dado un directorio src abre el directorio apuntado por el en dest
// Si src y dest son el mismo se cierra primero y se abre en su lugar
fs_dir* fs_opendir(fs_info* info, fs_dir* src, fs_dir* dest);
// Cierra el directorio
int64_t fs_closedir(fs_info* info, fs_dir* dir);
// Lee la proxima entrada en el directorio y la pone en su dir_entry
dir_entry* fs_readdir(fs_info* info, fs_dir* dir);
// Retrocede el directorio al comienzo como luego de un fs_opendir
fs_dir* fs_rewinddir(fs_info* info, fs_dir* dir);
// Crea un nuevo archivo
int64_t fs_create(fs_info* info, fs_dir* dir, dir_entry* file_info);
// Trunca un archivo al tamaño length
int64_t fs_truncate(fs_info* info, fs_file* file, uint64_t length);
// Borra el archivo apuntado por dir
int64_t fs_remove(fs_info* info, fs_dir* dir);

#endif
