#ifndef __FS_EXT2_H
#define __FS_EXT2_H

#include <types.h>

/* Formato del superblock de EXT2 como esta en OSDev
 * http://wiki.osdev.org/Ext2
 *
 * Todos los valores de tiempo estan en POSIX TIME.
 * Para campos con flags o valores especificos asociados
 * revisar los defines que aparecen luego del struct
*/

// Número mágico que identifica el filesystem
#define EXT2_MAGIC                      0xEF53
// El superbloque siempre mide 1024 bytes
#define EXT2_SUPERBLOCK_SIZE            1024 
// El superbloque siempre esta en offset 1024
#define EXT2_SUPERBLOCK_OFFSET          1024
// El superbloque es el indice 1 con bloques de 1024
#define EXT2_SUPERBLOCK_INDEX           1
// Los tamaños de bloque son shiftear este numero N veces a la izquierda
#define EXT2_MIN_BLOCK_SIZE             1024
// 1024 es 2 elevado a la 10
#define EXT2_MIN_BLOCK_SHIFT            10
// Los tamaños de fragmento son shiftear este numero N veces a la izquierda
#define EXT2_MIN_FRAGMENT_SIZE          1024

// Inodos reservados
#define EXT2_BAD_BLOCKS_INODE           1
#define EXT2_ROOT_INODE                 2

// Superbloque
typedef struct {
    uint32_t number_inodes;    // Cantidad de inodos
    uint32_t number_blocks;    // Cantidad de bloques
    uint32_t reserved_blocks;  // Cantidad de bloques reservados al superuser
    uint32_t free_blocks;      // Cantidad de bloques sin alloc
    uint32_t free_inodes;      // Cantidad de inodos sin alloc
    uint32_t bn_superblock;    // Numero de bloque del PRIMER superblock
    uint32_t block_size_shift; // Block size es 1024 shifteado por esto
    uint32_t frag_size_shift;  // Fragment size es 1024 shifteado por esto
    uint32_t blocks_per_group; // Cantidad de bloques en un block group
    uint32_t frags_per_group;  // Cantidad de fragmentos en un block group
    uint32_t inodes_per_group; // Cantidad de inodos en un block group
    uint32_t last_mount_time;  // Cuando fue la ultima vez que se montó 
    uint32_t last_write_time;  // Cuando fue la ultima vez que se escribió
    uint16_t mnts_since_check; // Mounts desde que se le hizo un checkeo
    uint16_t mnts_per_check;   // Mounts permitidos antes de hacerle un checkeo
    uint16_t ext2_magic;       // La firma de ext2, tiene que ser 0xEF53
    uint16_t fs_state;         // Estado del filesystem 
    uint16_t error_action;     // Que hacer en caso de error 
    uint16_t version_minor;    // La parte minor de la version
    uint32_t last_check_time;  // Cuando fue la ultima vez que se checkeó 
    uint32_t force_check_time; // El intervalo para que se fuerce un checkeo
    uint32_t os_id;            // ID del SO que lo creó 
    uint32_t version_major;    // La parte major de la version
    uint16_t priv_user_id;     // User ID que puede usar bloques reservados
    uint16_t priv_group_id;    // Group ID que puede usar bloques reservados
    uint32_t unreserved_inode; // Primer inodo sin reservar por el filesystem
    uint16_t inode_size;       // Tamaño del struct de un inodo en bytes
    uint16_t gn_superblock;    // Numero de block group del superblock
    uint32_t optional_flags;   // Features opcionales para usar el filesystem
    uint32_t req_flags;        // Features para usar el filesystem
    uint32_t req_write_flags;  // Features para escribir en el filesystem
    uint8_t  fs_id[16];        // ID del filesystem
    char     volume_name[16];  // Nombre del volumen
    char     last_mnt[64];     // Ruta del ultimo montaje del filesystem
    uint32_t comp_algorithms;  // Algoritmos de compresion usados
    uint8_t  blocks_for_files; // Cantidad de bloques a reservar por archivos
    uint8_t  blocks_for_dirs;  // Cantidad de bloques a reservar por directorios
    /*
    Como no vamos a soportar journaling lo dejamos aca 
    uint16_t unused            // Sin uso
    uint8_t  journal_id[16]    // ID del journal
    uint32_t journal_inode     // Inodo del journal
    uint32_t journal_device    // Device del journal
    uint32_t head_orphan       // Primer inodo de la lista de inodos huerfanos
    */
} __attribute__((__packed__)) ext2_superblock;

// Campo fs_state del superblock
#define EXT2_STATE_CLEAN                1
#define EXT2_STATE_ERROR                2

// Campo error_action del superblock
#define EXT2_ERROR_IGNORE               1
#define EXT2_ERROR_RO_REMOUNT           2
#define EXT2_ERROR_PANIC                3

// Campo os_id del superblock
#define EXT2_OS_ID_LINUX                0
#define EXT2_OS_ID_HURD                 1
#define EXT2_OS_ID_MASIX                2
#define EXT2_OS_ID_FREEBSD              3
#define EXT2_OS_ID_OTHER                4

// Campo optional_flags del superblock
#define EXT2_FEATURE_DIR_PREALLOC       1
#define EXT2_FEATURE_AFS                2
#define EXT2_FEATURE_HAS_JOURNAL        4
#define EXT2_FEATURE_EXTENDED_INODES    8
#define EXT2_FEATURE_RESIZE             16
#define EXT2_FEATURE_DIR_HASH           32

// Campo required_flags del superblock
#define EXT2_FEATURE_COMPRESSION        1
#define EXT2_FEATURE_DIR_TYPE           2
#define EXT2_FEATURE_JOURNAL_REPLAY     4
#define EXT2_FEATURE_JOURNAL_DEVICE     8

// Campo required_write_flags del superblock
#define EXT2_FEATURE_SPARSE             1
#define EXT2_FEATURE_64_BIT_FILE_SIZE   2
#define EXT2_FEATURE_DIR_BINARY_TREE    4

// Descriptor de grupos de bloque, cada entrada tiene un tamaño de 32 bytes
typedef struct {
    uint32_t bn_block_bitmap;  // Numero de bloque del bitmap de bloques
    uint32_t bn_inode_bitmap;  // Numero de bloque del bitmap de inodos
    uint32_t bn_inode_table;   // Numero de bloque de la tabla de inodos
    uint16_t free_blocks;      // Cantidad de bloques sin alloc
    uint16_t free_inodes;      // Cantidad de inodos sin alloc
    uint16_t dir_inodes;       // Cantidad de inodos usados para directorios
    uint16_t reserved[7];      // Reservado para padding a 32 bytes
} __attribute__((__packed__)) ext2_group_desc;

/*
 * Por convencion los SO guardan estos datos 
 * adentro del inodo (No forma parte de EXT2)
*/
typedef struct {
    uint8_t  number_fragment;  // Fragment Number 
    uint8_t  size_fragment;    // Fragment Size
    uint16_t type_mode_high;   // Parte alta del type access (HURD ONLY)
    uint16_t user_id_high;     // Parte alta del User ID asociado
    uint16_t group_id_high;    // Parte alta del Group ID asociado
    uint32_t user_id_author;   // User ID del file author (HURD ONLY)
} __attribute__((__packed__)) ext2_os_struct;


// El inodo, este struct tiene un tamaño de 128 bytes
typedef struct {
    uint16_t mode;             // Permisos flags y tipo del archivo
    uint16_t user_id;          // User ID asociado
    uint32_t size_low;         // Parte baja del tamaño en bytes del inodo
    uint32_t last_access_time; // Ultimo tiempo de acceso 
    uint32_t creation_time;    // Cuando fue creado
    uint32_t last_mod_time;    // Ultima vez modificado 
    uint32_t deletion_time;    // Cuando fue borrado 
    uint16_t group_id;         // Group ID asociado
    uint16_t hard_links;       // Cantidad de hard links que apuntan al inodo
    uint32_t used_sectors;     // Cantidad de sectores en uso por este inodo
    uint32_t flags;            // Flags del inodo
    uint32_t os_value;         // Usado por el SO
    uint32_t bn_direct[12];    // Punteros a bloque directos
    uint32_t bn_single;        // Puntero a bloque indirecto simple
    uint32_t bn_double;        // Puntero a bloque indirecto doble
    uint32_t bn_triple;        // Puntero a bloque indirecto triple
    uint32_t generation;       // Version del archivo, usado por NFS
    uint32_t bn_acl;           // Bloque con los atributos extendidos (ACL)
    uint32_t size_high;        // Parte alta del tamaño del inodo (Solo files)
    uint32_t bn_fragment;      // Bloque con el file fragment (OBSOLETO)
    ext2_os_struct os_struct;  // Usado por el SO
} __attribute__((__packed__)) ext2_inode;

/* Ninguna de estas entradas puede abarcar más de un bloque de datos, 
 * incluyendo el string del nombre.
 */
typedef struct {
    uint32_t inode_number;     // Inodo asociado al archivo 
    uint16_t rec_len;          // Desplazamiento hasta la próxima entrada
    uint8_t  name_len;         // Cantidad de bytes que ocupa el nombre
    uint8_t  file_type;        // Tipo de archivo
    /* Aquí empieza el string del nombre 
     * a menos que name_len sea cero. 
     * Usar ext2_dir_struct *p + 1 para obtener la posicion del string
     */
} __attribute__((__packed__)) ext2_dir_entry;

// Campo flags en el inodo
#define EXT2_INODE_ACCESS_STICKY_BIT    0x1
#define EXT2_INODE_ACCESS_SET_GROUP_ID  0x2
#define EXT2_INODE_ACCESS_SET_USER_ID   0x4

// Campo type en el inodo
#define EXT2_INODE_TYPE_FIFO            0x1
#define EXT2_INODE_TYPE_CHAR_DEVICE     0x2
#define EXT2_INODE_TYPE_DIRECTORY       0x4
#define EXT2_INODE_TYPE_BLOCK_DEVICE    0x6
#define EXT2_INODE_TYPE_REGULAR_FILE    0x8
#define EXT2_INODE_TYPE_SYMBOLIC_LINK   0xA
#define EXT2_INODE_TYPE_UNIX_SOCKET     0xC

// Campo flags en el inodo
#define EXT2_INODE_FLAG_SECURE_DELETION 0x00000001
#define EXT2_INODE_FLAG_RECORD          0x00000002
#define EXT2_INODE_FLAG_COMPRESSED      0x00000004
#define EXT2_INODE_FLAG_SYNC_UPDATE     0x00000008
#define EXT2_INODE_FLAG_IMMUTABLE       0x00000010
#define EXT2_INODE_FLAG_APPEND_ONLY     0x00000020
#define EXT2_INODE_FLAG_NO_DUMP         0x00000040
#define EXT2_INODE_FLAG_NO_ATIME_UPDATE 0x00000080
#define EXT2_INODE_FLAG_DIRTY           0x00000100
#define EXT2_INODE_FLAG_COMPRESSED_BLCK 0x00000200
#define EXT2_INODE_FLAG_NO_UNCOMPRESS   0x00000400
#define EXT2_INODE_FLAG_COMPRESSED_ERR  0x00000800
#define EXT2_INODE_FLAG_DIR_HASH        0x00001000
#define EXT2_INODE_FLAG_AFS             0x00002000
#define EXT2_INODE_FLAG_JOURNAL_DATA    0x00004000

#endif
