%include "defines.mac"

global protected_mode
global common_code

;; LINKER
extern grub_load_address
extern end_section_data
extern end_section_bss

;; GDT
extern GDT_DESC

;; LONG MODE
extern long_mode

; Header multiboot de GRUB
MAGIC       equ 0x1BADB002        ; 'Magic number' lets bootloader find the header
MODULEALIGN equ 1<<0              ; Align loaded modules on page boundaries
MEMINFO     equ 1<<1              ; Provide memory map
HEADERVALID equ 1<<16             ; Use this header instead of ELF
FLAGS       equ MODULEALIGN | MEMINFO | HEADERVALID 
CHECKSUM    equ -(MAGIC + FLAGS)

; Address section
HEADERADDR  equ grub_load_address ; Address of this header
LOADADDR    equ grub_load_address ; Start of the OS image
LOADEND     equ end_section_data  ; End of the OS image
BSSEND      equ end_section_bss   ; End of the BSS section
ENTRYADDR   equ protected_mode    ; Entry address

section .header
	multiboot_header: 
		dd MAGIC
		dd FLAGS
		dd CHECKSUM
        dd HEADERADDR
        dd LOADADDR
        dd LOADEND
        dd BSSEND
        dd ENTRYADDR

section .text

;;
;; Seccion de datos
;; -------------------------------------------------------------------------- ;;

msg_long_mode_err: db '[BSP] FAIL! 64 bits mode unavailable! Kernel Halted.', 0
msg_cpuid_err:     db '[BSP] FAIL! CPUID unavailable! Kernel Halted.', 0
msg_paging:        db '[BSP] Configuring PAE, identity mapping 4GiB... ', 0
msg_long_mode:     db '[BSP] Starting up in long mode... ', 0
msg_done:          db 'Done!', 10, 0

;----------------------------------------------------------------------------
;------------------------ Comienzo modo protegido ---------------------------
;----------------------------------------------------------------------------

BITS 32

; ======== CODIGO BSP ========
; Aca arranca el BSP despues de GRUB 
; http://www.gnu.org/software/grub/manual/multiboot/multiboot.html#Machine-state
protected_mode:
    ; Recargar la GDT, vengo del entorno de grub donde la GDT puede ser basura
    lgdt [GDT_DESC]

    ; Necesito cargar el segmento de datos de kernel
    mov ax, GDT_DESC_DATA_DPL0
    mov ds, ax
    mov es, ax
    mov fs, ax
    mov gs, ax
    mov ss, ax

    .clear_screen:
    ; Limpio la pantalla
    ; Formato fondo negro, letra blanca y ASCII de espacio
    mov ax, 0x0720 
    xor ecx, ecx
    .clear_loop:
        mov [ecx + VIDEO_ADDRESS], ax
        add ecx, 2
        cmp ecx, VIDEO_SIZE
        jnz .clear_loop
    ; Uso ebp para guardar datos de pantalla
    ; Color blanco de letra
    ; Fila 0 y Columna 0
    mov ebp, 0x00000F00

    .set_stack:
    ; Setear la pila para el BSP temporalmente
    ; La pila del BSP se encuentra en la primera posición del arreglo.
    mov esp, KERNEL_STACK_BSP

    .check_cpuid:
    ; Chequeo de disponibilidad de uso de CPUID
    ; http://wiki.osdev.org/Setting_Up_Long_Mode#Detection_of_CPUID
    pushfd               ; Store the FLAGS-register.
    pop eax              ; Restore the A-register.
    mov ecx, eax         ; Set the C-register to the A-register.
    xor eax, 1 << 21     ; Flip the ID-bit, which is bit 21.
    push eax             ; Store the A-register.
    popfd                ; Restore the FLAGS-register.
    pushfd               ; Store the FLAGS-register.
    pop eax              ; Restore the A-register.
    push ecx             ; Store the C-register.
    popfd                ; Restore the FLAGS-register.
    xor eax, ecx         ; Do a XOR-operation on the A-register and the C-register.
    jz no_cpuid          ; The zero flag is set, no CPUID.
    ; Aca estamos seguros que CPUID esta habilitado

    .check_long_mode:
    ; Deteccion de modo 64 bits y mensaje de error sino esta disponible halteamos.
    mov eax, 0x80000000  ; Pasamos parametros 0x80000000.
    cpuid
    cmp eax, 0x80000001  ; 0x80000001 significa que esta habilitado.
    jb no_long_mode      ; Si es menor, modo largo no esta disponible
    ; Aca tenemos certeza de que tenemos modo de 64 bits disponible

    .init_paging:
    ; Inicializo paginacion 4k
    mov esi, msg_paging
    call print_text
    ; Estoy usando paginacion IA-32e => bits CR0.PG=1 y CR4.PAE=1 y EFER.LME=1
    ; Mapeo 4GiB con paginas de 4KiB

    ; Limpiamos edx lo usamos como la parte alta de las entradas
    xor edx, edx
    
    ; Creamos solo una entrada en la PML4T que apunta a la PDPT
    mov edi, KERNEL_PML4T
    ; La PDPT esta 1 pagina despues y le ponemos los atributos 
    lea eax, [edi + PAGE_SIZE + PAGE_ATTRIBUTES]
    mov ecx, 1
    call fill_page_table

    ; Creamos solo 4 entradas en la PDPT que apuntan a PDTs
    ; Cada entrada mapea un GiB de memoria, en total 4GiB
    shl ecx, 2
    call fill_page_table

    ; Crear 4 * 512 entradas en las PDTs que apuntan a PTTs
    ; Se mapean 4 Page Directories completos
    shl ecx, 9
    call fill_page_table

    ; Crear 4 * 512 * 512 entradas en PT que apuntan a paginas de 4KiB
    ; Se mapean 4 * 512 Page Tables completas haciendo identity mapping
    and eax, PAGE_ATTRIBUTES
    shl ecx, 9
    call fill_page_table

    mov esi, msg_done
    call print_text

    ; Comienzo a inicializar 64 bits
    mov esi, msg_long_mode
    call print_text

; ======== CODIGO COMUN ========
; Aca arrancan los AP y el BSP sigue ejecutando
common_code:
    ; Apunto cr3 al PML4
    mov eax, KERNEL_PML4T
    mov cr3, eax
    ; Prender el bit 5 (Sexto bit) para activar PAE
    mov eax, cr4
    or eax, 1 << 5
    mov cr4, eax
    ; Activamos modo largo
    mov ecx, 0xC0000080 ; Seleccionamos EFER MSR poniendo 0xC0000080 en ECX
    rdmsr               ; Leemos el registro en EDX:EAX.
    or eax, 1 << 8      ; Seteamos el bit de modo largo.
    wrmsr               ; Escribimos nuevamente al registro.
    ; Activamos paginacion 
    mov eax, cr0        ; Obtenemos el registro de control 0 actual.
    or eax, 1 << 31     ; Habilitamos el bit de Paginacion.
    mov cr0, eax        ; Escribimos el nuevo valor sobre el registro de control
    ; Estamos en modo ia32e compatibilidad con 32 bits
    ; Ahora saltamos a modo largo, que tiene otra entrada en la GDT
    jmp GDT_DESC_CODE_DPL0_64:long_mode

; == Funciones auxiliares en 32 bits ==

; Llena una tabla de pagina de cualquier nivel
; Deja los registros bien posicionados para sucesivas llamadas
; edx:eax: La pagina a apuntar por la primera entrada con atributos
; ecx: La cantidad de entradas a llenar
; edi: La tabla a llenar
; clobbers: esi
fill_page_table:
    lea esi, [ecx * 8]
    add esi, edi
    .loop:
        mov [edi], eax
        mov [edi + 4], edx
        ; Incremento 4KiB
        add eax, PAGE_SIZE 
        ; Una entrada son 8 bytes
        add edi, 8
        cmp edi, esi
        jnz .loop
    ; Llenamos el resto de la tabla con ceros
    .loop_zero:
        test edi, PAGE_SIZE - 1
        jz .end
        mov dword [edi], 0
        add edi, 4
        jmp .loop_zero
    .end:
    ret

; Imprime texto en pantalla
; esi: Puntero al texto
; ebp: col|row|format
; clobbers: eax, ecx, dx
print_text:
    xor eax, eax
    mov ecx, ebp
    ; Formato
    mov dx, cx
    shr ecx, 16
    mov al, VIDEO_WIDTH * 2
    ; Fila
    mul cl
    add eax, VIDEO_ADDRESS
    ; Columna
    shr cx, 8
    .loop:
        mov dl, [esi]
        cmp dl, 0
        jz .end
        add esi, 1
        cmp dl, 10
        jz .new_line
        mov [eax + ecx * 2], dx
        add cx, 1
        cmp cx, VIDEO_WIDTH
        jnz .loop
    .new_line:
        ; Fila + 1
        add ebp, 0x10000
        add eax, VIDEO_WIDTH * 2
        xor ecx, ecx
        jmp .loop 
    .end:
    ; Restauramos la columna
    and ebp, 0xFFFFFF
    shl ecx, 24
    or ebp, ecx
    ret 

no_cpuid:
    mov esi, msg_cpuid_err
    mov bp, 0x0C00
    call print_text
    jmp halt_loop 

no_long_mode:
    mov esi, msg_long_mode_err
    mov bp, 0x0C00
    call print_text

halt_loop:
    xchg bx, bx
    hlt
    jmp halt_loop
