#include <defines.h>
#include <exception.h>
#include <kmalloc.h>

#define KHEAP_BSIZE(b) (((uintptr_t)(b)->next - (uintptr_t)(b)))

kheap_list_header* kheap_head = NULL;

void kheap_init() {
    // La direccion donde empieza el heap
    kheap_head = (kheap_list_header*)KERNEL_HEAP_START;
    // Donde comienzan los bloques del heap
    kheap_head->start =
        (kheap_header*)(KERNEL_HEAP_START + sizeof(kheap_list_header));
    // El final del heap donde esta el puntero (Dummy Block) que marca el limite
    kheap_head->end = (kheap_header*)(KERNEL_HEAP_END - sizeof(kheap_header*));
    // Un puntero a alguno de los elementos de la lista circular del heap
    kheap_head->free_ptr = kheap_head->start;
    // Hacemos que el bloque libre apunte al bloque dummy del final
    kheap_head->free_ptr->next = kheap_head->end;
    // Hacemos que ese bloque dummy apunte al bloque libre
    kheap_head->end->next = kheap_head->free_ptr;
    // Hacemos que el bloque libre se apunte a si mismo
    kheap_head->free_ptr->next_free = kheap_head->free_ptr;
}

// Devuelve el bloque libre previo al bloque libre donde encaja mejor
static kheap_header* best_fit_prev(kheap_list_header* klh, uint64_t size) {
    // Empezamos por el puntero que tenemos a la lista circular y su siguiente
    kheap_header* free_ptr_prev = klh->free_ptr;
    // Si es NULL entonces no tenemos espacio
    if (free_ptr_prev == NULL)
        return NULL;
    kheap_header* free_ptr = free_ptr_prev->next_free;
    // De entrada no encontramos nada
    uint64_t best_size = 0;
    kheap_header* best_prev = NULL;
    // Paramos cuando volvemos al mismo del comienzo
    kheap_header* stop = klh->free_ptr;
    do {
        // Averiguo el tamaño del bloque libre
        uint64_t free_size = KHEAP_BSIZE(free_ptr);
        // Si encontramos uno que encaja y
        // es mejor o no teniamos ninguno lo asignamos
        if (free_size >= size && (free_size < best_size || !best_prev)) {
            best_prev = free_ptr_prev;
            best_size = free_size;
        }
        free_ptr_prev = free_ptr;
        free_ptr = free_ptr->next_free;
    } while (free_ptr != stop);
    // Si no encontramos nada vamos a devolver NULL
    return best_prev;
}

/* Asignamos algun lugar en el heap del kernel
 * con la cantidad de memoria que nos pidieron
 */
static void* kheap_alloc(kheap_list_header* klh, uint64_t size) {
    // Si nos piden size 0 algo anda mal
    if (size == 0)
        KERNEL_PANIC("Se pidio un kmalloc de 0");
    // Necesitamos al espacio para el puntero al proximo libre si queda libre
    if (size < sizeof(kheap_header*))
        size = sizeof(kheap_header*);
    // Necesitamos pedir espacio para el puntero que apunta al siguiente tambien
    size += sizeof(kheap_header*);
    // Buscar el bloque libre donde encaje mejor
    kheap_header* best_prev = best_fit_prev(klh, size);
    // Si no hay mas espacio, panic
    if (best_prev == NULL)
        KERNEL_PANIC("No hay espacio para kmalloc");
    // Nos dieron el anterior al mejor, obtenemos el mejor
    kheap_header* best = best_prev->next_free;
    // Al bloque que vamos a devolver lo inicializamos al principio
    kheap_header* new_alloc = best;
    // Calculamos cuanto nos sobra de espacio
    uint64_t remainder = KHEAP_BSIZE(best) - size;
    // Los espacios libres tienen que por lo menos tener espacio para el header
    if (remainder < sizeof(kheap_header)) {
        // No sobra espacio para el bloque libre, damos todo el bloque al alloc
        if (best == best_prev) {
            // Es el ultimo bloque libre entonces ponemos free_ptr en null
            klh->free_ptr = NULL;
        } else {
            // No es el ultimo, contamos el anterior con el siguiente
            best_prev->next_free = best->next_free;
            // Cambiamos el puntero de entrada a la lista a best_prev
            // Es para mejorar las chances en la busqueda, no es necesario
            klh->free_ptr = best_prev;
        }
    } else {
        // Lo ponemos al final del bloque libre
        new_alloc = (kheap_header*)((uintptr_t)new_alloc + remainder);
        // El bloque ocupado ahora va a apuntar al bloque que estaba adelante
        new_alloc->next = best->next;
        // Lo que sobro apunta al bloque nuevo
        best->next = new_alloc;
    }
    // Devolvemos el bloque menos el puntero del principio
    return (void*)((uintptr_t)new_alloc + sizeof(kheap_header*));
}

uint64_t kheap_available(kheap_list_header* klh) {
    kheap_header* free_ptr = klh->free_ptr;
    if (!free_ptr)
        return 0;
    uint64_t total = 0;
    // Sumamos todo el espacio libre para alloc
    do {
        total += KHEAP_BSIZE(free_ptr);
        free_ptr = free_ptr->next_free;
    } while (free_ptr != klh->free_ptr);
    return total;
}

static void kheap_free(kheap_list_header* klh, void* p) {
    // Este es el puntero que tenemos que buscar
    kheap_header* ptr = (kheap_header*)((uintptr_t)p - sizeof(kheap_header*));
    // Si esta fuera de los limites del heap hay algo mal
    if (ptr < klh->start || ptr >= klh->end)
        KERNEL_PANIC("Se quizo liberar memoria fuera del heap");
    // Empezamos por el puntero que tenemos a la lista circular
    kheap_header* free_ptr_prev = klh->free_ptr;
    // Necesitamos encontrar desde donde hasta donde tenemos que buscar
    kheap_header* start = NULL;
    kheap_header* end = NULL;
    if (free_ptr_prev) {
        // Tomamos el siguiente para comparar
        kheap_header* free_ptr = free_ptr_prev->next_free;
        kheap_header* stop = free_ptr;
        do {
            // O esta entre dos bloques libres
            // O esta despues del ultimo bloque libre o antes del primero
            if ((free_ptr_prev < ptr && ptr < free_ptr) ||
                (free_ptr <= free_ptr_prev &&
                 (free_ptr_prev < ptr || ptr < free_ptr)))
                break;
            // Avanzamos
            free_ptr_prev = free_ptr;
            free_ptr = free_ptr->next_free;
        } while (free_ptr != stop);
        start = free_ptr_prev->next;
        end = free_ptr;
    } else {
        // Si estamos llenos empezamos del comienzo al final
        start = klh->start;
        end = klh->end;
    }
    // Nuestro iterador
    kheap_header* block;
    // Buscamos el bloque entre start y end
    for (block = start; block != ptr && block != end; block = block->next);
    // Si llegamos a end no lo encontramos, panic
    if (block == end)
        KERNEL_PANIC("Se quizo liberar memoria invalida");
    // Si llegamos aca es porque lo encontramos, tenemos 5 casos
    if (free_ptr_prev) {
        if (block == start) {
            // Caso 1: El bloque esta solo entre los bloques libres
            // Nos queda un bloque libre que es la fusion de los 3 bloques
            // Caso 2 (Else): El bloque tiene antes un bloque libre
            // Fusionamos el bloque con el anterior
            if (block->next == end)
                *free_ptr_prev = *end;
            else
                free_ptr_prev->next = block->next;
        } else {
            // Si el que encontramos no esta al principio
            // Tenemos que hacer que el anterior libre apunte a nosotros
            free_ptr_prev->next_free = block;
            // Caso 3: El bloque tiene despues un bloque libre
            // Fusionamos el bloque con el siguiente
            // Caso 4 (Else): El bloque no tiene bloques libres en los extremos
            // Lo "agregamos" a la lista de libres
            if (block->next == end)
                *block = *end;
            else
                block->next_free = end;
        }
        // Necesario, el siguiente libre al previo es el unico que quedo valido
        klh->free_ptr = free_ptr_prev->next_free;
    } else {
        // Caso 5: Estabamos llenos, ahora pasamos a tener un bloque libre
        block->next_free = block;
        klh->free_ptr = block;
    }
}

void* kmalloc(uint64_t size) {
    return kheap_alloc(kheap_head, size);
}

void kfree(void* mem) {
    if (!mem)
        return;
    kheap_free(kheap_head, mem);
}
