#include <defines.h>
#include <elf.h>
#include <loader.h>
#include <utils.h>

static const char elf_magic[] = {0x7F, 'E', 'L', 'F'};

application_data app_data;

// Retorna 0 si está todo bien.
static int64_t elf_read_header(elf_header *header) {
    // Leemos el header y chequeamos que esté todo en orden
    if (memcmp(header->magic, elf_magic, sizeof(elf_magic)) ||
        header->magic[EI_CLASS] != ELFCLASS64 ||
        header->magic[EI_DATAENC] != ELFDATA2LSB)
        return INVALID_ELF;
    return 0;
}

// Pega la imagen dada por el archivo elf en memoria.
static int64_t overlay_image(int64_t fd, elf_header *elf) {
    // Ojo con esto: la direccion contra la cual fue linkeado el binario ELF
    // tiene que estar mapeada en delirios (y disponible físicamente).
    // Y de paso, convendría asegurarse que no esté pisando nada del kernel
    // (ver mapa de memoria en defines.h).
    // elf->program_header = (elf_pheader*) (imagep+header->ph_offset);
    elf_pheader program_header;
    uint64_t entry_count = elf->ph_entry_count;
    for (uint64_t i = 0; i < entry_count; i++) {
        // Lo que queremos es el header de una cierta sección del elf.
        // Antes que nada tenemos que computar el offset donde se encuentra.
        // En general, esto es:
        // (POS_INICIAL + elf->ph_offset + sizeof(elf_pheader) * i) bytes
        // POS_INICIAL es el comienzo de la imagen elf (siempre 0 en este caso).
        seek(fd, elf->ph_offset + sizeof(elf_pheader) * i, SEEK_SET);
        if (!read(fd, &program_header, sizeof(elf_pheader)))
            return ERRREADING_ELF;
        if (program_header.type == ELF_LOAD && program_header.memory_size > 0) {
            uint64_t filesize = program_header.file_size;
            uintptr_t address = program_header.virtual_address;
            seek(fd, program_header.offset, SEEK_SET);
            if (!read(fd, (void *)address, filesize))
                return ERRREADING_ELF;
            // Inicializamos a cero el resto (section .bss/COMMON)
            memset((void *)(address + filesize), 0,
                   program_header.memory_size - filesize);
        }
    }
    return 0;
}

static int64_t read_and_load(int64_t fd, uint64_t *entry_point) {
    elf_header header;
    if (!read(fd, &header, sizeof(elf_header)))
        return ERRREADING_ELF;
    int64_t check = elf_read_header(&header);
    if (check)
        return check;
    *entry_point = header.entry_point;
    return overlay_image(fd, &header);
}

// Nota: El archivo recibido no es cerrado hasta que termina la aplicación
//       y vuelve el flujo de ejecución al shell.
application_data *load_exec_elf64(int64_t fd, int argc, char *argv[]) {
    // Cargamos el código y datos en memoria.
    if (!(app_data.loader_error = read_and_load(fd, &app_data.entry_point))) {
        // Acá viene la parte peliaguda:
        // Hay que guardar la información del stack en app_data.
        // Puede que lo mejor sea hacer lo que sigue en ensamblador.
        execute_start(argc, argv);
        // Eventualmente el flujo de ejecución vuelve y el shell se encarga
        //  de imprimir en pantalla la información interesante.
    }
    return &app_data;
}
