#include <core_id.h>
#include <exception.h>
#include <i386.h>
#include <idt.h>
#include <ioapic.h>
#include <kernel.h>
#include <lapic.h>
#include <multicore_boot.h>
#include <ports.h>
#include <timer.h>
#include <utils.h>

// Signature de los headers
static const char MPFS_SIG[] = {'_', 'M', 'P', '_'};
static const char MPCT_SIG[] = {'P', 'C', 'M', 'P'};

static processor_entry processors[MAX_PROCESSORS];
static uint64_t processor_count = 0;
// Indice del bootstrap processor
static int64_t bsp_index = -1;

static uint16_t get_ebda(void) {
    // De acuerdo a OSDEV, y en GRUB parece asumir lo mismo
    // http://linux4u.jinr.ru/pub/misc/sys/grub/SVN/grub/grub2/commands/i386/pc/acpi.c
    // El start de la EBDA puede en la mayoria de los casos tomarse de la
    // posicion 0x40E, o usando BIOS INT 12.
    // TODO: Averiguar si GRUB ya te da este valor.
    return (*((uint16_t *)0x40e)) << 4;
}

static bool check_valid_mpfs(const mp_float_struct *mpfs) {
    if (memcmp(mpfs->signature, MPFS_SIG, sizeof(MPFS_SIG)))
        return false;
    // La suma tiene que dar 0
    return byte_checksum(mpfs, mpfs->length * 16) == 0;
}

// Chequea el checksum de la tabla de configuracion de Multi Processor
static bool check_valid_mpct(const mp_config_table *mpct) {
    if (memcmp(mpct->signature, MPCT_SIG, sizeof(MPCT_SIG)))
        return false;
    // La suma tiene que dar 0
    return byte_checksum(mpct, mpct->length) == 0;
}

static mp_float_struct *find_mpfs_in_region(void *region, uint64_t len) {
    uint8_t *start = (uint8_t *)region;
    uint8_t *end = start + len;
    for (uint8_t *p = start; p < end; p++) {
        if (memcmp(p, MPFS_SIG, sizeof(MPFS_SIG)) == 0) {
            mp_float_struct *mpfs = (mp_float_struct *)p;
            if (check_valid_mpfs(mpfs))
                return mpfs;
        }
    }
    return NULL;
}

static mp_float_struct *find_mpfs(void) {
    mp_float_struct *mpfs;
    // La ebda mide al menos un kilobyte segun
    // http://www.nondot.org/sabre/os/files/Booting/BIOS_SEG.txt
    // Solo nos interesa el primer KB por la MPSpec.
    if ((mpfs = find_mpfs_in_region((void *)(uintptr_t)get_ebda(), 0x400)))
        return mpfs;
    // El ultimo KB de memoria base. Asumimos que 640 KB es el minimo.
    // Ignorando que obviamente eso debiera alcanzarle a cualquiera.
    if ((mpfs = find_mpfs_in_region((void *)(639 << 10), 0x400)))
        return mpfs;
    // Direcciones de BIOS ROM
    if ((mpfs = find_mpfs_in_region((void *)0xF0000, 0x10000)))
        return mpfs;
    return NULL;
}

// Procesa la entrada de procesador para configurar este core
static void configure_processor(const processor_entry *entry) {
    // Agregar el procesador a la lista de procesadores.
    PANIC_IF(processor_count == MAX_PROCESSORS);
    memcpy(processors + processor_count, entry, sizeof(processor_entry));
    if (entry->bootstrap)
        bsp_index = processor_count;
    processor_count++;
}

// Determinar posicion de la proxima entrada de la tabla
static const mp_entry *next_mp_entry(const mp_config_table *mpct,
                                     const mp_entry *p) {
    PANIC_UNLESS(p->entry_type < MP_ENTRY_TYPE_COUNT);
    // Solo la entrada PROCESSOR tiene un tamaño diferente
    if (p->entry_type == MP_ENTRY_PROCESSOR)
        return (const mp_entry *)((uintptr_t)p + MP_ENTRY_PROCESSOR_SIZE);
    return (const mp_entry *)((uintptr_t)p + MP_ENTRY_DEFAULT_SIZE);
}

// Inicializa a modo APIC el IMCR del BSP
static void start_icmr_apic_mode(void) {
    // TODO: Conseguir maquina donde probar esto
    PANIC_UNLESS(false);
}

// Levanta la posicion de memoria a donde los aps van a saltar
static void set_warm_reset_vector(uint32_t address) {
    // Seteamos al BIOS para que el shutdown sea reset warm por jump
    cmos_writeb(0xF, 0xA);
    uint32_t segment = (address << 12) & 0xFFFF0000;
    uint32_t offset = address & 0xFF;
    // Le ponemos a que direccion saltar en la posicion 40:67h en modo real.
    *((uint32_t *)((0x40 << 4) + 0x67)) = segment | offset;
}

// Enciende todos los APs
static void turn_on_aps(uint32_t ap_start) {
    // Copiamos el código antes de prender los aps
    uint64_t ap_code_size =
        (uint64_t)(((uintptr_t)&ap_real_mode_end) - ((uintptr_t)&ap_real_mode));
    memcpy((void *)(uint64_t)ap_start, &ap_real_mode, ap_code_size);
    // De acuerdo a http://www.cheesecake.org/sac/smp.html
    // Primero hay que poner Warm Reset with far jump en el CMOS
    // y poner la direccion a la que va a saltar en ese lugar.
    // Adicionalmente hay que limpiar los errores del registro de local APIC
    set_warm_reset_vector(ap_start);
    lapic_clear_errors();

    // Las STARTUP Ipis se envian solamente si el local APIC no es un 82489DX.
    bool send_startup_ipis = !lapic_is_82489();
    for (uint64_t i = 0; i < processor_count; i++) {
        processor_entry *p = processors + i;
        // No despertar al BSP, porque no es AP
        if (p->bootstrap)
            continue;
        // Crear mensaje de INIT, INIT Deasserted y STARTUP para IPI
        intr_command_register init_ipi, init_ipi_doff, startup_ipi;
        // Inicializarlos
        lapic_init_ipi(&init_ipi, INIT, 0, p->lapic_id, 0);
        lapic_init_ipi(&init_ipi_doff, INIT_DEASSERT, 0, p->lapic_id, 0);
        // En vector para startup ipi hay que enviar la pagina fisica
        // donde va a empezar a ejecutar. Para eso hay que shiftear 12 bits.
        // El valor del startup IPI ya se considera alineado a pagina.
        lapic_init_ipi(&startup_ipi, STARTUP, ap_start >> 12, p->lapic_id, 0);
        // Enviar las ipis de inicio
        lapic_send_ipi(&init_ipi);
        lapic_wait_ipi();
        lapic_send_ipi(&init_ipi_doff);
        lapic_wait_ipi();
        // Dormir un poco mas de 10 milisegundos
        sleep(11);
        if (send_startup_ipis) {
            lapic_clear_errors();
            // Enviar las STARTUP ipis, dormir y esperar.
            lapic_send_ipi(&startup_ipi);
            // Dormir un poco mas de 20 milisegundos
            sleep(21);
            lapic_wait_ipi();
            lapic_send_ipi(&startup_ipi);
            // Dormir un poco mas de 20 milisegundos
            sleep(21);
            lapic_wait_ipi();
        }
    }
}

// Determina la configuracion del procesador si hay una tabla de configuracion
// Es decir si no hay una configuracion default en uso.
static uint32_t determine_cpu_config(const mp_float_struct *mpfs) {
    uint32_t ioapic_address = 0;
    // Chequeamos la tabla de configuracion
    mp_config_table *mpct = (mp_config_table *)(uintptr_t)mpfs->config;
    if (!check_valid_mpct(mpct))
        KERNEL_PANIC("Invalid MultiProcessor Configuration Table");
    // Seguimos las entradas de la tabla de configuracion
    PANIC_IF(mpct->entry_count == 0);
    const mp_entry *entry = mpct->entries;
    for (uint64_t i = 0; i < mpct->entry_count; i++) {
        if (entry->entry_type == MP_ENTRY_PROCESSOR)
            configure_processor(&entry->chunk.processor);
        // TODO: Que pasa si hay mas de un IOAPIC activo?
        // Por ahora seleccionamos uno solo
        if (!ioapic_address && entry->entry_type == MP_ENTRY_IOAPIC &&
            entry->chunk.ioapic.enabled)
            // Devolvemos la direccion del ioapic
            ioapic_address = entry->chunk.ioapic.base_address;
        // Por ahora ignoramos todas las demas entradas.
        entry = next_mp_entry(mpct, entry);
    }
    // Si no encontramos bootstrap processor estamos en un serio problema.
    // Porque no hay chance de que podamos prender el resto de los cores.
    // TODO: Puede suceder esto?
    if (bsp_index < 0)
        KERNEL_PANIC("Bootstrap processor entry not found");
    // Seteamos la direccion donde esta mapeada la local apic
    lapic_set_address(mpct->local_apic_addr);
    // Si ICMR presente, hay que levantarlo a modo APIC
    if (mpfs->mp_features2 & IMCRP_BIT)
        start_icmr_apic_mode();
    // Devolvemos la direccion del ioapic
    return ioapic_address;
}

static void determine_default_config(mp_float_struct *mpfs) {
    // TODO: Conseguir maquina donde probar esto, determinar si es necesario.
    PANIC_UNLESS(false);
}

static void core_id_init() {
    // Al agregar primero al bsp, su core id es 0.
    core_id_add(processors[bsp_index].lapic_id);
    for (uint32_t i = 0; i < processor_count; i++) {
        if (i == (uint32_t)bsp_index)
            continue;
        core_id_add(processors[i].lapic_id);
    }
}

void multiprocessor_init() {
    // Detectar MPFS
    mp_float_struct *mpfs = find_mpfs();
    if (!mpfs)
        KERNEL_PANIC("Estructura MPFS no encontrada");
    uint32_t ioapic_address = DEFAULT_IOAPIC_ADDR;
    // Recorrer estructura y determinar los cores
    if (mpfs->config) {
        // Hay que determinar la configuracion
        ioapic_address = determine_cpu_config(mpfs);
    } else if (mpfs->mp_features1) {
        // La configuracion ya esta definida y es una estandar
        determine_default_config(mpfs);
    } else {
        KERNEL_PANIC("Estructura MPFS invalida");
    }
    // Hay que enmascarar los dos PIC, vamos a usar IOAPIC
    mask_pic();
    // Somos el BSP prendemos nuestro local APIC
    lapic_turn_on();
    // Armamos un diccionario para identificar los cores a partir de su apic id.
    core_id_init();
    // Inicializamos el ioapic que encontramos, siempre hay al menos uno
    initialize_ioapic(ioapic_address, processors[bsp_index].lapic_id);
    // Inicializamos el RTC para poder hacer sleep
    timer_init();
    // Activamos interrupciones
    sti();
    // Prendemos los AP en la posicion predefinida debajo del primer MB
    turn_on_aps(AP_STARTUP_ADDRESS);
}
