#include <command.h>
#include <defines.h>
#include <exception.h>
#include <io.h>
#include <print.h>
#include <shell.h>
#include <utils.h>

const shell_command command_list[] = {{"cd", change_directory},
                                      {"ls", list_directory},
                                      {"echo", echo_string},
                                      {"exec", load_binary}};

const uint64_t command_count = sizeof(command_list) / sizeof(command_list[0]);

char line_buffer[SHELL_MAX_LEN];
char cwd[SHELL_MAX_LEN];
uint64_t cwd_length;

int64_t check_path(char* path) {
    return (type(path) == FILE_TYPE_DIRECTORY) - 1;
}

char* get_cwd() {
    return &cwd[0];
}

int64_t set_cwd(char* new_cwd) {
    uint64_t len = strlen(new_cwd);
    if (len >= SHELL_MAX_LEN) {
        print("Path demasiado largo\n");
        return -1;
    }
    if (check_path(new_cwd)) {
        print("Directorio invalido o inexistente\n");
        return -1;
    }
    memcpy(cwd, new_cwd, len + 1);
    cwd_length = len;
    // Dejamos siempre la barra del final si no esta
    if (cwd[cwd_length - 1] != '/') {
        cwd[cwd_length] = '/';
        cwd[cwd_length + 1] = 0;
        cwd_length++;
    }
    // Actualizo el working directory
    set_cwd_dir(cwd);
    return 0;
}

int64_t set_relative_cwd(char* new_relative) {
    uint64_t len = cwd_length + strlen(new_relative);
    if (cwd_length + strlen(new_relative) >= SHELL_MAX_LEN) {
        print("Path relativo demasiado largo\n");
        return -1;
    }
    if (check_path(new_relative)) {
        print("Directorio invalido o inexistente\n");
        return -1;
    }
    strcat(cwd, new_relative);
    cwd_length = len;
    // Dejamos siempre la barra del final si no esta
    if (cwd[cwd_length - 1] != '/') {
        cwd[cwd_length] = '/';
        cwd[cwd_length + 1] = 0;
        cwd_length++;
    }
    // Actualizo el working directory
    set_cwd_dir(new_relative);
    return 0;
}

int64_t readline(char* buffer, uint64_t len) {
    uint64_t read_bytes = read(STDIN, buffer, len);
    // Si leyo demasiado devolvemos un error
    if (read_bytes >= len)
        return -1;
    // Marcamos el final del string
    buffer[read_bytes] = '\0';
    return read_bytes;
}

int64_t execute_command(char* command, uint64_t argc) {
    char* argv[argc];
    argv[0] = command;
    // Asignamos a argv cada uno de los parametros restantes
    for (uint64_t i = 0, j = 1; j < argc; i++) {
        if (command[i] == 0) {
            argv[j] = &command[i + 1];
            j++;
        }
    }
    for (uint64_t i = 0; i < command_count; i++) {
        if (strcmp(command_list[i].name, command) == 0)
            return command_list[i].command(argc, argv);
    }
    print("No se reconocio el comando\n");
    return 0;
}

int64_t parse_command(char* command, uint64_t length) {
    // El final del string
    char* end = command + length;
    // Punteros iteradores del string
    char* i = command;
    char* j = command;
    // Ignoramos los espacios del comienzo
    while (j < end && *j == ' ')
        j++;
    // Si ya llegamos al final es porque no habia nada
    if (j >= end)
        return 0;
    // Tenemos al menos un parametro
    uint64_t argc = 1;
    while (j < end) {
        // Tomamos el caracter y avanzamos
        char c = *j++;
        switch (c) {
            // Ignorar todos los espacios que sobran entre parametros
            case ' ':
                // Ponemos un 0 para terminar el parametro
                *i++ = '\0';
                // Ignoramos todos los espacios que siguen
                while (j < end && *j == ' ')
                    j++;
                // Si no llegamos al final encontramos otro parametro
                if (j < end)
                    argc++;
                break;
            // Apostrofes delimitan strings
            case '\'':
                // Metemos todo hasta que encontramos otro apostrofe
                while (j < end) {
                    if (*j == '\'') {
                        j++;
                        break;
                    }
                    *i++ = *j++;
                }
                break;
            // Barra invertida, representamos literal el siguiente caracter
            case '\\':
                if (j < end)
                    *i++ = *j++;
                break;
            // Caracter no especial, lo escribimos
            default:
                *i++ = c;
        }
    }
    // Escribimos un 0 en el final del ultimo parametro
    *i = '\0';
    return argc;
}

noreturn void shell_init() {
    if (set_cwd("/") < 0)
        KERNEL_PANIC("No se pudo abrir el root");
    while (true) {
        printf("DeliriOS:%s# ", get_cwd());
        int64_t read_chars = readline(line_buffer, SHELL_MAX_LEN);
        // Leimos demasiado
        if (read_chars < 0) {
            printf("\nExpresion demasiado larga\n");
            continue;
        }
        int64_t argc = parse_command(line_buffer, read_chars);
        if (argc <= 0)
            continue;
        int64_t err_code = execute_command(line_buffer, argc);
        if (err_code)
            printf("El comando devolvio codigo de error: %i\n", err_code);
    }
}
