#include <exception.h>
#include <io.h>
#include <macros.h>
#include <screen.h>
#include <utils.h>

static fd_entry fd_table[MAX_FILE_DESCRIPTORS];
static dd_entry dd_table[MAX_DIR_DESCRIPTORS];
static fs_dir cwd_dir;
fs_info fs;

static int64_t get_free_fd() {
    for (int64_t i = 0; i < MAX_FILE_DESCRIPTORS; i++)
        if (!fd_table[i].stat.mode)
            return i;
    return -1;
}

static int64_t get_free_dd() {
    for (int64_t i = 0; i < MAX_DIR_DESCRIPTORS; i++)
        if (!dd_table[i].valid)
            return i;
    return -1;
}

static fd_entry *get_fd_entry(int64_t fd) {
    if (fd < 0 || fd > MAX_FILE_DESCRIPTORS || !fd_table[fd].stat.mode)
        return NULL;
    return fd_table + fd;
}

static dd_entry *get_dd_entry(int64_t dd) {
    if (dd < 0 || dd > MAX_DIR_DESCRIPTORS || !dd_table[dd].valid)
        return NULL;
    return dd_table + dd;
}

static bool is_root(const char *path) {
    return (path[0] == '/' && path[1] == 0);
}

static bool is_valid_type(uint8_t type) {
    return (type == FILE_TYPE_FIFO || type == FILE_TYPE_CHAR_DEVICE ||
            type == FILE_TYPE_DIRECTORY || type == FILE_TYPE_BLOCK_DEVICE ||
            type == FILE_TYPE_REGULAR_FILE || type == FILE_TYPE_SYMBOLIC_LINK ||
            type == FILE_TYPE_UNIX_SOCKET);
}

/** Busca una entrada delimitada por start y end en fs_dir
 *  Retorna dir y apunta a la entrada si se encuentra
 *  Retorna NULL y dir queda invalido sino
 */
static dir_entry *seek_dir_entry(const char *start, const char *end,
                                 fs_dir *dir) {
    uint64_t length = end - start;
    if (length >= MAX_FILENAME_SIZE)
        return NULL;
    dir_entry *entry;
    while ((entry = fs_readdir(&fs, dir)))
        if (strncmp(entry->name, start, length) == 0 && !entry->name[length])
            break;
    return entry;
}

/** Clona el directorio actual en dir
 *  Retorna dir si se pudo concretar o NULL si no
 */
static fs_dir *clone_cwd(fs_dir *dir) {
    const char *same_dir = ".";
    fs_rewinddir(&fs, &cwd_dir);
    seek_dir_entry(same_dir, same_dir + 1, &cwd_dir);
    return fs_opendir(&fs, &cwd_dir, dir);
}

/** Busca la siguiente barra
 *  Devuelve un puntero a dicha barra o al final del string
 */
static const char *find_next_slash(const char *name) {
    while (*name && *name != '/')
        name++;
    return name;
}

/** Abre sucesivos directorios del path en dir, excepto el ultimo nombre
 *  Retorna la posicion del ultimo nombre
 *  Si el path es relativo, empieza desde cwd. Si no, desde el root.
 *  Si se pudo concretar el recorrido,
 *  dir quedará abierto en el anteultimo nombre del path
 *  Si no se pudo, dir permancerá cerrado
 */
static const char *traverse_path(const char *path, fs_dir *dir) {
    // Si es un path absoluto arrancamos desde el root
    if (*path == '/') {
        // La barra del comienzo la ignoramos
        path++;
        fs_root(&fs, dir);
    } else {
        // Soporte para paths relativos
        // Con esto dir es un fs_file independiente del cwd.
        clone_cwd(dir);
    }
    const char *last_name;
    while (*(last_name = path)) {
        path = find_next_slash(path);
        // Si no encontramos barra o la encontramos al final
        if (!path[0] || !path[1])
            break;
        if (!seek_dir_entry(last_name, path, dir)) {
            fs_closedir(&fs, dir);
            return NULL;
        }
        if (!fs_opendir(&fs, dir, dir))
            return NULL;
        path++;
    }
    return last_name;
}

/** Combina traverse_path y avanza a la entrada del ultimo elemento del path
 *  Si se pudo concretar se retorna la dir_entry apuntandolo
 *  Si no permanecerá cerrado retornando NULL
 */
static dir_entry *get_dir_entry(const char *path, fs_dir *dir) {
    const char *last = traverse_path(path, dir);
    if (!last)
        return NULL;
    dir_entry *entry = seek_dir_entry(last, find_next_slash(last), dir);
    if (!entry)
        fs_closedir(&fs, dir);
    return entry;
}

/** Combina get_dir_entry y abre la entrada de directorio correctamente
 *  Si se pudo concretar se retorna dir con el directorio abierto
 *  Si no permanecera cerrado y retornara NULL
 */
static fs_dir *get_dir(const char *path, fs_dir *dir) {
    if (is_root(path))
        return fs_root(&fs, dir);
    // Sino recorremos el filesystem
    dir_entry *entry = get_dir_entry(path, dir);
    if (!entry)
        return NULL;
    if (entry->file_type != FILE_TYPE_DIRECTORY) {
        fs_closedir(&fs, dir);
        return NULL;
    }
    return fs_opendir(&fs, dir, dir);
}

/** Inicializa el filesystem y los descriptores de pantalla
 */
int64_t io_init() {
    if (fs_mount(&fs))
        KERNEL_PANIC("Error al montar el filesystem");
    if (!fs_root(&fs, &cwd_dir))
        KERNEL_PANIC("Error al abrir el directorio root.");
    // Tapamos los file descriptors standard
    fd_table[STDIN].stat.mode = 0xFF;
    fd_table[STDOUT].stat.mode = 0xFF;
    fd_table[STDERR].stat.mode = 0xFF;
    return 0;
}

int64_t io_kill() {
    fs_unmount(&fs);
    return 0;
}

int64_t set_cwd_dir(const char *path) {
    if (path == NULL || path[0] == '\0')
        return -1;
    fs_dir new_wd;
    if (!get_dir(path, &new_wd))
        return -1;
    // Todo anduvo bien! Modifico el working directory actual.
    fs_closedir(&fs, &cwd_dir);
    cwd_dir = new_wd;
    return 0;
}

int64_t read(int64_t fd, void *buf, uint64_t count) {
    if (fd == STDIN)
        return screen_read(buf, count);
    if (fd <= STDERR)
        return -1;
    fd_entry *entry = get_fd_entry(fd);
    if (!entry)
        return -1;
    // Cuanto podemos leer antes de llegar al final?
    uint64_t remaining = entry->stat.size - fs_tell(&fs, &entry->file);
    return fs_read(&fs, &entry->file, buf, MIN(count, remaining));
}

int64_t write(int64_t fd, const void *buf, uint64_t count) {
    if (fd <= STDIN)
        return -1;
    if (fd == STDOUT)
        return screen_write(buf, count);
    if (fd == STDERR)
        return screen_write_err(buf, count);
    // Conseguimos la entry del file
    fd_entry *entry = get_fd_entry(fd);
    if (!entry)
        return -1;
    uint64_t written = fs_write(&fs, &entry->file, buf, count);
    fs_fstat(&fs, &entry->file, &entry->stat);
    return written;
}

/**
 * Devuelve la posición absoluta en el archivo file.
 */
int64_t tell(int64_t fd) {
    fd_entry *entry = get_fd_entry(fd);
    if (!entry)
        return -1;
    return fs_tell(&fs, &entry->file);
}

/**
 * Coloca el iterador fs_file en la posición origin+offset bytes.
 * @param offset El offset en bytes, puede ser negativo,
 *               si se da uno invalido la funcion no hace nada
 * @param origin La posición absoluta desde donde se computa el offset.
 */
int64_t seek(int64_t fd, int64_t offset, int64_t origin) {
    int64_t relative = 0;
    fd_entry *entry = get_fd_entry(fd);
    if (!entry)
        return -1;
    // Tanto si es SEEK_END como SEEK_SET restamos la posicion actual
    switch (origin) {
        case SEEK_CUR:
            break;
        case SEEK_END:
            relative += entry->stat.size;
        case SEEK_SET:
            relative -= fs_tell(&fs, &entry->file);
            break;
        default:
            return -1;
    }
    return fs_seek(&fs, &entry->file, relative + offset);
}

int64_t truncate(int64_t fd, uint64_t length) {
    fd_entry *entry = get_fd_entry(fd);
    if (!entry)
        return -1;
    if (length > entry->stat.size) {
        length -= entry->stat.size;
        char zeros[1024] = {0};
        while (length >= 1024) {
            fs_write(&fs, &entry->file, zeros, 1024);
            length -= 1024;
        }
        if (length)
            fs_write(&fs, &entry->file, zeros, length);
    } else if (length < entry->stat.size) {
        if (fs_truncate(&fs, &entry->file, length))
            return -1;
    }
    fs_fstat(&fs, &entry->file, &entry->stat);
    return 0;
}

uint8_t type(const char *path) {
    if (is_root(path))
        return FILE_TYPE_DIRECTORY;
    fs_dir dir;
    dir_entry *entry = get_dir_entry(path, &dir);
    if (!entry)
        return 0;
    uint8_t type = entry->file_type;
    fs_closedir(&fs, &dir);
    return type;
}

int64_t create(const char *path, uint8_t filetype) {
    // No podes el root ni un tipo no definido
    if (is_root(path) || !is_valid_type(filetype))
        return -1;
    fs_dir dir;
    const char *filename = traverse_path(path, &dir);
    if (!filename)
        return -1;
    const char *end = find_next_slash(filename);
    uint64_t name_length = end - filename;
    // Si el path es un directorio y el filetype no corresponde
    // O el nombre es demasiado largo
    // O el archivo existe, retornamos error
    if ((*end && filetype != FILE_TYPE_DIRECTORY) ||
        name_length >= MAX_FILENAME_SIZE ||
        seek_dir_entry(filename, end, &dir)) {
        fs_closedir(&fs, &dir);
        return -1;
    }
    fs_rewinddir(&fs, &dir);
    // Creamos la entrada con el nombre y el tipo
    dir_entry entry;
    memcpy(entry.name, filename, name_length);
    entry.name[name_length] = '\0';
    entry.file_type = filetype;
    // Creamos el archivo
    int64_t err = fs_create(&fs, &dir, &entry);
    fs_closedir(&fs, &dir);
    return err;
}

int64_t remove(const char *path) {
    fs_dir dir;
    if (!get_dir_entry(path, &dir))
        return -1;
    int64_t err = fs_remove(&fs, &dir);
    fs_closedir(&fs, &dir);
    return err;
}

int64_t open(const char *path) {
    int64_t fd = get_free_fd();
    if (fd < 0)
        KERNEL_PANIC("Maximum file descriptors reached");
    fd_entry *entry = fd_table + fd;
    fs_dir dir;
    if (!get_dir_entry(path, &dir))
        return -1;
    fs_open(&fs, &dir, &entry->file);
    // Actualizamos el estado del archivo
    fs_fstat(&fs, &entry->file, &entry->stat);
    fs_closedir(&fs, &dir);
    return fd;
}

int64_t close(int64_t fd) {
    fd_entry *entry = get_fd_entry(fd);
    if (!entry)
        return -1;
    fs_close(&fs, &entry->file);
    // Invalido el archivo
    entry->stat.mode = 0;
    return 0;
}

int64_t opendir(const char *path) {
    int64_t dd = get_free_dd();
    if (dd < 0)
        KERNEL_PANIC("Maximum directory descriptors reached");
    dd_entry *entry = dd_table + dd;
    if (!get_dir(path, &entry->dir))
        return -1;
    entry->valid = 1;
    return dd;
}

int64_t closedir(int64_t dd) {
    dd_entry *entry = get_dd_entry(dd);
    if (!entry)
        return -1;
    fs_closedir(&fs, &entry->dir);
    // Invalido el directorio
    entry->valid = 0;
    return 0;
}

dir_entry *readdir(int64_t dd) {
    dd_entry *entry = get_dd_entry(dd);
    if (!entry)
        return NULL;
    return fs_readdir(&fs, &entry->dir);
}

int64_t rewinddir(int64_t dd) {
    dd_entry *entry = get_dd_entry(dd);
    if (!entry)
        return -1;
    if (!fs_rewinddir(&fs, &entry->dir))
        return -1;
    return 0;
}

/**
 * Devuelve si se llego al EOF o no
 */
uint64_t eof(int64_t fd) {
    fd_entry *entry = fd_table + fd;
    return entry->stat.size == (uint64_t)fs_tell(&fs, &entry->file);
}
