#include <macros.h>
#include <types.h>
#include <utils.h>

#define is_clear(x, bit) (((x) & (1 << (7 - bit))) == 0)
#define is_set(x, bit) (((x) & (1 << (7 - bit))) != 0)

// Estos me dan bytes con n bits seteados al principio/final, y el resto en 0.
// Start quiere decir los mas significativos, end los menos significativos
#define N_BIT_START(n) ((1 << (n)) - 1)
#define N_BIT_END(n) (N_BIT_START(n) << (8 - (n)))

// Count leading zeros, espera un numero de 8 bits
uint64_t clz8(int8_t byte) {
    for (uint8_t i = 0; i < 7; i++)
        if (byte & ((1 << 7) >> i))
            return i;
    return 8;
}

// Count trailing zeros, espera un numero de 8 bits
uint64_t ctz8(uint8_t byte) {
    for (uint8_t i = 0; i < 7; i++)
        if (byte & (1 << i))
            return i;
    return 8;
}

/* @brief Cuenta la cantidad de ceros consecutivos en el bitmap
 *        desde offset hasta size. WARNING: size no es multiplo de 8
 *        (raro), puede haber invalid reads.
 *
 * @param _bitmap El bitmap en cuestion.
 * @param size El tamano del bitmap en bits.
 * @param offset Desde que bit empezar a buscar ceros.
 *
 * @return Cantidad de ceros desde offset
 */

uint64_t bitmap_count_zeros(const void* const _bitmap, uint64_t size,
                            uint64_t offset) {
    // TODO: Posible optimizacion? leer todo en 32 bits.
    // Contras: invalid reads, endianness :-(
    uint64_t bit_count = 0;
    uint8_t* bitmap = (uint8_t*)_bitmap;
    uint64_t size_copy = size;
    uint64_t offset_copy = offset;
    // Avanzo mientras no tenga que setear nada
    while (offset > 7) {
        offset -= 8;
        bitmap++;
        size -= 8;
    }
    uint8_t byte_without_trailing = (*bitmap) >> offset;
    if (byte_without_trailing != (uint8_t)0) {
        return ctz8(byte_without_trailing);
    } else {
        bit_count += 8 - offset;
        bitmap++;
        size -= 8;
    }
    while (*bitmap == (uint8_t)0) {
        bit_count += 8;
        bitmap++;
        size -= 8;
    }
    bit_count += ctz8(*bitmap);
    return MIN(size_copy - offset_copy, bit_count);
}

/* @brief Busca /count/ cantidad de bits en cero en el bitmap.
 *
 * @param _bitmap El bitmap en cuestion.
 * @param size El tamano del bitmap en bits.
 * @param offset Desde que bit empezar a buscar ceros.
 * @param count Cuantos ceros buscar.
 *
 * @return Posicion a partir de la cual encontro los bits. Si es size,
 *         significa que no encontro.
 */

uint64_t bitmap_search_zeros(const void* _bitmap, uint64_t size,
                             uint64_t offset, uint64_t count) {
    uint64_t current = offset;
    while (current < size) {
        // Busco ceros al principio desde _bitmap + cur / 8, dado que cur/8
        // es la cantidad de bytes que me adelante desde el inicio del bitmap
        // leading_zeros es en realidad trailing_zeros
        uint64_t leading_zeros = bitmap_count_zeros(_bitmap, size, current);
        if (leading_zeros >= count)
            return current;
        else
            // + 1 porque seguro el siguiente es un 1, asi que lo salteamos
            current += leading_zeros + 1;
    }
    return current;
}

/* @brief Modifica el bitmap seteandole count cantidad de bits
 *        de offset en adelante.
 *
 * @param _bitmap El bitmap en cuestion.
 * @param size El tamaño del bitmap (en bits).
 * @param offset El tamaño en bits del bitmap en cuestion.
 * @param count Cuantos bits queremos setear.
 *
 * @return 0 -> salio todo bien
 *         1 -> se paso de size
 *         2 -> ya estaban seteados (probablemente este corrupto el fs)
 */
uint64_t bitmap_set_bits(const void* _bitmap, uint64_t size, uint64_t offset,
                         uint64_t count) {
    if (offset + count > size)
        return 1;
    uint8_t* bitmap = (uint8_t*)_bitmap;
    // Avanzo mientras no tenga que setear nada
    while (offset > 7) {
        offset -= 8;
        bitmap++;
        size -= 8;
    }
    // Tengo un byte y le tengo que setear cierta cantidad de bits del final
    // por ejemplo, si en este momento offset es 0, le tengo que setear todos
    // los bits. Si offset es 7, le tengo que setear el último.
    if (count < 8 - offset) {
        uint8_t bitmask = N_BIT_END(8 - offset) & N_BIT_START(offset + count);
        *bitmap |= bitmask;
        count = 0;
    } else {
        *bitmap |= N_BIT_END(8 - offset);
        count -= 8 - offset;
        bitmap++;
        size -= 8;
    }
    while (count > 7) {  // Mientras tenga que setear 8 bits o mas
        if (*bitmap != (uint8_t)0)
            return 2;
        *bitmap |= ~((uint8_t)0);
        count -= 8;
        bitmap++;
        size -= 8;
    }
    // Tengo un byte y le tengo que setear cierta cantidad de bits del principio
    // por ejemplo, si en este momento count es 0, no le tengo que setear nada
    // Si offset es 7, los 7 bits mas significativos.
    if (count > 0)
        *bitmap |= N_BIT_START(count);
    return 0;
}

/* @brief Modifica el bitmap poniendo en 0 count cantidad de bits
 *        de offset en adelante.
 *
 * @param _bitmap El bitmap en cuestion.
 * @param size El tamaño del bitmap (en bits).
 * @param offset El tamaño en bits del bitmap en cuestion.
 * @param count Cuantos bits queremos setear.
 *
 * @return 0 -> salio todo bien
 *         1 -> se paso de size
 *         2 -> ya estaban en 0 (probablemente este corrupto el fs)
 */
uint64_t bitmap_clear_bits(const void* _bitmap, uint64_t size, uint64_t offset,
                           uint64_t count) {
    if (offset + count > size)
        return 1;
    uint8_t* bitmap = (uint8_t*)_bitmap;
    // Avanzo mientras no tenga que setear nada
    while (offset > 7) {
        offset -= 8;
        bitmap++;
        size -= 8;
    }
    // Tengo un byte y le tengo que setear cierta cantidad de bits del final
    // por ejemplo, si en este momento offset es 0, le tengo que setear todos
    // los bits. Si offset es 7, le tengo que setear el último.
    if (count < 8 - offset) {
        uint8_t bitmask = N_BIT_END(8 - offset) & N_BIT_START(offset + count);
        *bitmap &= ~bitmask;
        count = 0;
    } else {
        *bitmap &= N_BIT_START(offset);
        count -= 8 - offset;
        bitmap++;
        size -= 8;
    }
    while (count > 7) {  // Mientras tenga que clearear 8 bits o mas
        *bitmap &= ((uint8_t)0);
        count -= 8;
        bitmap++;
        size -= 8;
    }
    // Tengo un byte y le tengo que setear cierta cantidad de bits del principio
    // por ejemplo, si en este momento count es 0, no le tengo que setear nada
    // Si offset es 7, los 7 bits mas significativos.
    if (count > 0)
        *bitmap &= N_BIT_END(8 - count);

    return 0;
}

void byte_string(uint8_t byte, char* buf) {
    for (uint64_t i = 0; i < 8; i++) {
        buf[7 - i] = (byte % 2) + '0';
        byte /= 2;
    }
}

void bitmap_string(const void* _bitmap, char* buffer, uint64_t size) {
    uint8_t* bitmap = (uint8_t*)_bitmap;
    char* buf = buffer;
    for (uint64_t i = 0; i < size / 8; i++) {
        byte_string(*bitmap, buf);
        bitmap++;
        buf += 8;
    }
    buf[0] = '\0';
}
